# -*- coding: utf-8 -*-
# Licensed under GPLv3+ - see LICENSE
from __future__ import annotations
from collections import abc
from typing import Optional, Union, Iterable, Sequence, Self, Generator
import configparser
from importlib import resources
import numpy as np
from rich import print as rprint
from astropy import units as u
from astropy import coordinates as coord
from astropy.time import Time
from astroplan import Observer, FixedTarget, is_observable, is_event_observable, is_always_observable
from dataclasses import dataclass
from enum import Enum, auto
from . import constraints
from . import freqsetups

"""Module that defines the `Station` and `Stations` objects, which represent a station (antenna)
or a network composed of antennas.
"""

__all__: list[str] = ["Station", "Stations", "Mount", "MountType"]


class MountType(Enum):
    ALTAZ = auto()
    EQUAT = auto()


@dataclass
class Axis:
    """Defines one of the mount axis for a given antenna.
    The axis is defined by the angle range it can move (limits), and the slewing speed and
    acceleration.
    """
    limits: tuple[u.Quantity, u.Quantity]
    speed: u.Quantity
    acceleration: u.Quantity = u.Quantity(0.0, u.deg / u.s / u.s)


@dataclass
class Mount:
    """Defines a telescope mount, including the mount type (ALTAZ, EQUAT), the limits of the
    telescope to point in the sky, and the slewing velocity (rate) and acceleration (acc).
    """
    mount_type: MountType
    ax1: Axis
    ax2: Axis


class Station(object):
    """Defines an astronomical station (antenna).
    A station is defined by some names and codenames, coordinates, and its sensitivity for the
    radio bands (wavelengths) it can observe.
    Apart of the metadata related to the station, it allows to compute the altitude/azimuth,
    elevation, or simply when a source is visible from the station for a given time range.
    """

    def __init__(self, name: str, codename: str, networks: tuple[str], location: coord.EarthLocation,
                 freqs_sefds: dict[str, u.Quantity], fullname: Optional[str] = None,
                 country: str = "", diameter: str = "",
                 real_time: bool = False, mount: Optional[Mount] = None,
                 max_datarate: Optional[u.Quantity | dict[str, u.Quantity]] = None) -> None:
        """Initializes a station.

        Inputs
        - name : str
            Name of the station.
            If it contains undercores (_), they will be converted to blank spaces.
        - codename : str
            A short code (accronym) for the name of the station. It is meant to follow the standard
            approach from the EVN: an (often) two-letter code unique for each station.
        - networks : tuple[str]
            Name of the network(s) to which the station belongs (e.g. EVN).
        - location : astropy.coordinates.EarthLocation
            Position of the station on Earth in (x,y,z) gecentric coordinates.
        - freqs_sefds : dict
            Dictionary with all frequencies the station can observe as keys of the dictionary, and
            the values representing the system equivalent flux density (SEFD; in Jansky units)
            at each frequency.
            Although the key format is in principle free, we recommend to use the syntax 'XXcm' (str type).
            This will be then consistent with the default station catalog and will avoid issues for some
            functions.
        - fullname : str [OPTIONAL]
            Full name of the station. If not given, same as `name` is assumed.
            It can be used to expand the full name if an abbreviation is typically used for the name.
            For example, name: VLA, fullname: Karl G. Jansky Very Large Array.
        - country : str [OPTIONAL]
            Country where the station is located.
        - diameter : str [OPTIONAL]
            Diameter of the station (free format string). We recommend a syntax of e.g. '30 m' for normal
            single-dish antennas, and in case of interferometers it can have a form like '25 x 20 m',
            meaning that the station is composed of 25 antennas of 20 m each.
        - real_time : bool [OPTIONAL]
            If the station can participate in real-time observations (e.g. e-EVN), False by default.
        - mount : Mount [OPTIONAL]
            The mount of the station, including the type of mount and the slewing limits, speed,
            and acceleration.
            If not provided, it will assume an ALTAZ mount with no pointing limits and very high
            slewing speed.
        - max_datarate : u.Quantity or dict[str, u.Quantity]  [OPTIONAL]
            Specifies the maximum data rate that the station can record. If not provided, it will use the
            data rate assumed in the observation. It can be either a astropy.unit.Quantity value
            (equivalent to Mb/s), or a dictionary in the case that the antenna can show different data rates
            when it is participating in different networks. In that case, the abbreviation of the
            network will be the key of the dictionary, with the quantity as value.
            For example, Darnhall can record at 4 Gbps when participating within e-MERLIN observations,
            but their data rate is limited to 512 Mbps when participating within the EVN.
        """
        for a_var, a_var_name in zip((name, codename, country, diameter),
                                     ("name", "codename", "country", "diameter")):
            assert isinstance(a_var, str), f"'{a_var_name}' must be a str."

        assert isinstance(fullname, str) or fullname is None
        assert isinstance(real_time, bool), "'real_time' must be a bool."
        self.observer: Observer = Observer(name=name.replace("_", " "), location=location)
        self._codename: str = codename
        self._networks: tuple[str] = networks if isinstance(networks, tuple) else (networks,)
        self._freqs_sefds: dict[str, float] = {f if "cm" in f else f"{f}cm": v
                                               for f, v in freqs_sefds.items()}
        self._fullname: str = name if fullname is None else fullname
        self._country: str = country
        self._diameter: str = diameter
        self._real_time: bool = real_time
        if mount is None:
            self._mount: Mount = Mount(MountType.ALTAZ, Axis((-10*u.deg, 370*u.deg),
                                                             300*u.deg/u.s, 0.0*u.deg/u.s/u.s),
                                       Axis((10*u.deg, 100*u.deg), 300*u.deg/u.s, 0.0*u.deg/u.s/u.s))
        else:
            assert isinstance(mount, Mount)
            self._mount = mount

        self._max_datarate: u.Quantity | dict | None = max_datarate
        if self.mount.mount_type == MountType.ALTAZ:
            self._constraints = [constraints.AzimuthConstraint(min=self.mount.ax1.limits[0],
                                                               max=self.mount.ax1.limits[1]),
                                 constraints.ElevationConstraint(min=self.mount.ax2.limits[0],
                                                                 max=self.mount.ax2.limits[1])]
        else:
            self._constraints = [constraints.HourAngleConstraint(min=self.mount.ax1.limits[0],
                                                                 max=self.mount.ax1.limits[1]),
                                 constraints.DeclinationConstraint(min=self.mount.ax2.limits[0],
                                                                   max=self.mount.ax2.limits[1]),
                                 constraints.ElevationConstraint(min=5*u.deg)]

    @property
    def name(self) -> str:
        """Name of the station."""
        assert self.observer.name is not None
        return self.observer.name

    @property
    def codename(self) -> str:
        """Codename of the station (typically a two-letter accronym)."""
        return self._codename

    @property
    def fullname(self) -> str:
        """Full name of the station. If not specified, it can be the same as 'name'."""
        return self._fullname

    @property
    def networks(self) -> tuple[str]:
        """Name of the network(s) to which the station belongs."""
        return self._networks

    @property
    def country(self) -> str:
        """Country where this station is located.
        It can be an empty string if not specified.
        """
        return self._country

    @property
    def diameter(self) -> str:
        """String representing the diameter of the station, and/or how many antennas compose
        the station in case of connected-interferometers.
        """
        return self._diameter

    @property
    def real_time(self) -> bool:
        """If the station can participate in real-time observations (e.g. e-EVN)."""
        return self._real_time

    @property
    def location(self) -> coord.EarthLocation:
        """Location of the station as an astropy.coordinates.EarthLocation object."""
        return self.observer.location

    @property
    def constraints(self) -> list[constraints.Constraint]:
        """Returns the observing constraints for the station"""
        return self._constraints

    @property
    def bands(self) -> list[str]:
        """Observing bands the station can observe.
        Returns a dict_keys object with all bands in a string format as introduced in the freqs_sefd
        attribute when the Station was created.
        """
        return list(self._freqs_sefds.keys())

    @property
    def sefds(self) -> dict[str, u.Quantity]:
        """Returns a dictionary with the system equivalent flux density (SEFDs) for each
        of the frequencies the station can observe (given as keys).
        """
        return self._freqs_sefds

    @property
    def mount(self) -> Mount:
        """Returns the mount of the station"""
        return self._mount

    @property
    def max_datarate(self) -> Union[u.Quantity, dict[str, u.Quantity], None]:
        """Returns the maximum data rate that the station can observe.
        It can be either a astropy.unit.Quantity value (equivalent to Mb/s),
        or a dictionary in the case that the antenna can show different data rates
        when it is participating in different networks. In that case, the abbreviation of the network will
        be the key of the dictionary, with the quantity as value. For example, Darnhall can record at 4 Gbps
        when participating within e-MERLIN observations, but their data rate is limited to 512 Mbps when
        participating within the EVN.
        """
        return self._max_datarate

    def elevation(self, obs_times: Time, target: FixedTarget) -> coord.angles.Latitude:
        """Returns the elevation of the target source as seen by the Station during obs_times.

        Inputs
        - obs_times : astropy.time.Time
            Time to compute the elevation of the source
            (either a single timestamp or an array of times).
        - target : astroplan.FixedTarget
             Target to observe.

        Output
        - elevations : astropy.coordinates.angles.Latitute
            Elevation of the source at the given obs_times.
        """
        return self.observer.altaz(obs_times, target).alt

    def altaz(self, obs_times: Time, target: FixedTarget) -> coord.SkyCoord:
        """Returns the altaz coordinates of the target source for the given observing times.

        Inputs
        - obs_times : astropy.time.Time
            Time to compute the elevation of the source
            (either a single timestamp or an array of times).
        - target : astroplan.FixedTarget
             Target coordinates to observe.

        Output
        - altaz : astropy.coordinates.sky_coordinate.SkyCoord
            Altitude and azimuth of the source for each given time.
        """
        return self.observer.altaz(obs_times, target)

    def hour_angle(self, obs_times: Time, target: FixedTarget) -> coord.Angle:
        """Returns the local hour angle of the target at the given time."""
        return self.observer.target_hour_angle(time=obs_times, target=target)

    def is_observable(self, obs_times: Time, target: FixedTarget) -> list[bool]:
        """Returns when the target source is observable for this station at the given times.

        Inputs
        - obs_times : astropy.time.Time
            Time to compute the elevation of the source
            (either a single timestamp or an array of times).
        - target : astroplan.FixedTarget
             Target coordinates to observe. If None, the target would be assumed to be visible at all times.

        Output
        - visible : list[list[bool]]
            List of booleans of same length as targets for whether or not each target is ever
            observable in the time range given the constraints.
        """
        return is_event_observable(self.constraints, self.observer, target, times=obs_times)[0, :]

    def is_ever_observable(self, obs_times: Time, target: FixedTarget) -> list[bool]:
        """Returns if the target source is observable for this station at least for part of the times.

        Inputs
        - obs_times : astropy.time.Time
            Time to compute the elevation of the source
            (either a single timestamp or an array of times).
        - target : astroplan.FixedTarget
             Target coordinates to observe. If None, the target would be assumed to be visible at all times.

        Output
        - visible : list[bool]
            List of booleans of same length as targets for whether or not each target is ever
            observable in the time range given the constraints.
        """
        return is_observable(self.constraints, self.observer, target, times=obs_times)

    def is_always_observable(self, obs_times: Time, target: FixedTarget) -> bool:
        """Returns whenever the target source is observable for this station at all given times.

        Inputs
        - obs_times : astropy.time.Time
            Time to compute the elevation of the source
            (either a single timestamp or an array of times).
        - target : astroplan.FixedTarget
             Target coordinates to observe. If None, the target would be assumed to be visible at all times.

        Output
        - visible : bool
            Whether or not each target is observable in the time range given the constraints.
        """
        return is_always_observable(self.constraints, self.observer, target, times=obs_times)[0]

    def has_band(self, band: str) -> bool:
        """Returns if the Station can observed the given band `the_band`.

        Inputs
        - band : str
            A string representing an observing band, following the same syntax as used
            when the station was initialized and the bands where defined in the keys of
            the freqs_sefds attribute.

        Output
        - bool whenever the station has the given observing band.
        """
        return band in self.bands

    def sefd(self, band: str) -> u.Quantity:
        """Returns the system equivalent flux density (SEFD) of the Station at the given band,
        in Jansky (Jy) units.

        Input
        - band : str
            A string representing an observing band, following the same syntax as used
            when the station was initialized and the bands where defined in the keys of
            the freqs_sefds attribute.

        Output
        - SEFD : float
            The SEFD at the given band, in Jy units.

        Exception
        - It may raise KeyError if the given band is not available for this station.

        """
        return self._freqs_sefds[band]

    def slewing_time(self, coordinates1: coord.AltAz, coordinates2: coord.AltAz, time: Time) -> u.Quantity:
        # TODO: this is not the best implementation (asking for a time). Check later.
        """Returns the expected slewing time to move from coordinates1 to coordinates2 according to
        the known velocity and acceleration from the mount.

        Args
        - coordinates1 : astropy.coordinates.AltAz
            AltAz object with the coordinates of the first point. Must be in altaz coordinates.
        - coordinates2 : astropy.coordinates.AltAz
            AltAz object with the coordinates of the second point. Must be in altaz coordinates.

        Returns
        - slewing_time : astropy.time.TimeDelta
            TimeDelta object with the slewing time.
        """
        assert isinstance(coordinates1, coord.AltAz) and isinstance(coordinates2, coord.AltAz)
        # If the antenna mount is AltAz, then all the same; but if it's equatorial, then it needs to be RA/DEC.
        separation = coordinates1.spherical_offset_to(coordinates2)
        if self.mount.mount_type == MountType.ALTAZ:
            altaz1 = self.altaz(time, coordinates1)
            altaz2 = self.altaz(time, coordinates2)
            separation = (altaz1.az - altaz2.az, altaz1.alt - altaz2.alt)
        else:
            separation = coordinates1.spherical_offset_to(coordinates2)

        t = [0, 0]
        for i, axis in enumerate([self.mount.ax1, self.mount.ax2]):
            if axis.acceleration == 0.0:
                t[i] = separation[i]/axis.speed
            elif separation[i] <= 0.5*axis.speed**2/axis.acceleration:
                t[i] = np.sqrt(2*separation[i]/axis.acceleration)
            else:
                t[i] = separation[i]/axis.speed + 0.5*axis.speed/axis.acceleration

        return max(t)

    def __str__(self):
        return f"<{self.codename}>"

    def __repr__(self):
        return f"<Station: {self.codename}>"


class Stations(object):
    """Defines a network of antennas (collection) that can participate in an observation together.
    """

    def __init__(self, filename: Optional[str] = None, stations: Optional[Iterable[Station]] = None,
                 observing_bands: Optional[Sequence[str]] = None,
                 max_datarates: Optional[Union[Sequence[u.Quantity], u.Quantity]] = None) -> None:
        """Initializes a Stations of antennas.

        Inputs
        - filename : str  [OPTIONAL]
            Name of the file that contains the information of the antennas
            (by default, the 'stations_catalog.inp').
        - stations : list of Station-type elements  [OPTIONAL]
            List with all stations belonging to the given network. If not provided, it will include
            all antennas present in the catalogs.
        - observing_bands : list of str  [OPTIONAL]
            List with all possible observing bands, as specified in the freqsetups.py file.
            If not provided, it will consider all observing bands from the available antennas.
        - full_name : str  [OPTIONAL]
            Full name (expanded) associated to the network.
            If not provided, it will be name
        - max_datarates :  list of u.Quantity  or u.Quantity or None [OPTIONAL]
            Maximum data rates to be able to record at each observing band. If provided as list,
            it must have the same dimensions as `observing_bands`. If scalar, means that the
            network has the same maximum data rate at all bands.
            If not provided, it will assume no upper limit on the data rate.

        Raises
            ValueError: If 'max_datarates' is provided, 'observing_bands' must also be provided.
        """
        assert isinstance(stations, abc.Iterable) or stations is None, "'stations' must be a list or be empty."
        self._stations: dict[str, Station] = {}
        self._bands: dict[str, u.Quantity] = {}
        if (observing_bands is None) and (max_datarates is not None):
            raise ValueError("If 'max_datarates' is provided, 'observing_bands' must also be provided.")

        if observing_bands is not None:
            if isinstance(max_datarates, Sequence):
                assert len(observing_bands) == len(max_datarates), \
                       "'observing_bands' and 'max_datarates' must have the same dimensions."
                for i, band in enumerate(observing_bands):
                    self._bands[band] = max_datarates[i]
            elif max_datarates is not None:
                for band in observing_bands:
                    self._bands[band] = max_datarates
            else:
                for band in observing_bands:
                    self._bands[band] = None

        if stations is not None:
            for a_station in stations:
                assert isinstance(a_station, Station), \
                       f"There is an element in 'stations' that is not a Station object ({a_station})."
                if a_station.codename not in self._stations.keys():
                    self._stations[a_station.codename] = a_station
                else:
                    rprint(f"[yellow bold]WARNING: {a_station.codename} is duplicated in the provided "
                           "stations list while initializing a network.[/yellow bold]")

                if observing_bands is None:
                    for aband in a_station.bands:
                        self._bands[aband] = None
        else:
            # Read all stations from catalog,
            for station in self._get_stations_from_configfile(filename):
                self._stations[station.codename] = station
                for aband in station.bands:
                    self._bands[aband] = None
            # I don't think I need to know the networks at this point
            # _all_networks: dict[str, Stations] = self.get_networks_from_configfile(networks_filename)
            # TODO: implement this part!!!!!!!!!!!!
            # raise NotImplementedError

    @property
    def stations(self) -> list[Station]:
        """Returns a list containing all stations in the network."""
        return list(self._stations.values())

    @stations.setter
    def stations(self, new_stations: list[Station]):
        self._stations = {s.codename: s for s in new_stations}

    @property
    def number_of_stations(self) -> int:
        """Returns the total number of stations in the network."""
        return len(self.stations)

    @property
    def station_names(self) -> list[str]:
        """Returns the names from all the stations in the network"""
        return [s.name for s in self._stations.values()]

    @property
    def station_codenames(self) -> list[str]:
        """Returns a dict_keys with the `codenames` from all the stations in the network."""
        return list(self._stations.keys())

    @property
    def observing_bands(self) -> list[str]:
        """Returns a set with all `bands` that the stations in the network can observe,
        or at least a subset of stations.
        """
        return list(self._bands.keys())

    def max_datarate(self, observing_band: str) -> Optional[u.Quantity]:
        return self._bands[observing_band]

    def add_station(self, station: Union[Station, Iterable[Station]]):
        """Adds a new station to the network.
        If a station with the same codename is already present, it will do nothing.

        Inputs
        - station : Station
            Station to be added to the network.
        """
        assert isinstance(station, Station) or isinstance(station, Iterable), \
               "station must be a Station object or a Iterable of Station objects."
        if isinstance(station, Station):
            if station.codename in self.station_codenames:
                rprint(f"[yellow bold]WARNING: {station.codename} already in stations. "
                       "Ignoring addition.[/yellow bold]")
            else:
                self._stations[station.codename] = station
        else:
            for a_stat in station:
                if a_stat.codename in self.station_codenames:
                    rprint(f"[yellow bold]WARNING: {a_stat.codename} already in stations. "
                           "Ignoring addition.[/yellow bold]")
                else:
                    self._stations[a_stat.codename] = a_stat

    def remove_station(self, a_station: Station):
        """Removes a station from the network.
        If the station is not present, then it will do nothing.

        Inputs
        - a_station : Station
            Station to be added to the network.
        """
        if a_station.codename in self.station_codenames:
            self._stations.__delitem__(a_station.codename)

    def __str__(self):
        return f"<Stations ({self.number_of_stations}): <{', '.join(self.station_codenames)}>>"

    def __len__(self):
        return self._stations.__len__()

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._stations[self.station_codenames[key]]
        elif key in self._stations.keys():
            return self._stations[key]
        elif key in self.station_names:
            return self._stations[self.station_codenames[self.station_names.index(key)]]
        else:
            raise KeyError(f"Key {key} is not in the network.")

    def __setitem__(self, key, value):
        self._stations[key] = value
        self._codenames = tuple(self._stations.keys())

    def __delitem__(self, key):
        if isinstance(key, int):
            self._stations.__delitem__(self.keys[key])
        else:
            self._stations.__delitem__(key)

        self._codenames = tuple(self._stations.keys())

    def __iter__(self):
        return iter(self._stations.values())

    def __contains__(self, item):
        return self._stations.__contains__(item)

    def __add__(self, stations: Stations):
        if not isinstance(stations, Stations):
            raise TypeError("The element to add must be a Stations object")

        obs_bands = [b for b in self.observing_bands if b in stations.observing_bands]
        max_dt = []
        for b in obs_bands:
            if self.max_datarate(b) is None:
                max_dt.append(stations.max_datarate(b))
            elif stations.max_datarate(b) is None:
                max_dt.append(self.max_datarate(b))
            else:
                max_dt.append(min(self.max_datarate(b), stations.max_datarate(b)))  # type: ignore

        return Stations(stations=set(self.stations + stations.stations),
                        observing_bands=obs_bands, max_datarates=max_dt)

    def __iadd__(self, stations: Stations):
        return self.__add__(stations)

    def stations_with_band(self, band: str) -> Generator[Station]:
        """Returns a the stations in the Stations that can observe at the given band.
        that can observe at the given band.

        Inputs
        - band : str
            The observing band.

        Returns
        - stations : Generator[Station]
            A list containing only the stations that can observe at the given band.
        """
        for station in self.stations:
            if band in station.bands:
                yield station

    def filter_networks(self, networks: Union[str, Sequence[str]], only_defaults: bool = False) -> Stations:
        """Returns a new network which only contains the antennas that are part
        of the given network

        Input
            - networks : str or Sequence[str]
                The name of the network or networks to which all stations need to belong to.
            - only_defaults : bool  [DEFAULT = False]
                If True, it will only return the antennas that are part of the default
                network, ignoring all others.

        Returns
            - Stations
                A new network object containing only the antennas that can observe in the network.
        """
        # TODO: change this to network -> networks so it may include multiple!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        if isinstance(networks, str):
            networks = [n.strip() for n in networks.split(',')]

        if only_defaults:
            all_networks = self.get_networks_from_configfile()
            if len(networks) > 0:
                the_network = all_networks[networks[0]]
                for a_network in networks[1:]:
                    the_network += all_networks[a_network]
            else:
                raise ValueError("networks cannot be an empty sequence!")

            bands, datarates = None, None
            for aband, adatarate in zip(self.observing_bands, self._bands.values()):
                if aband in the_network.observing_bands:
                    if bands is None:
                        bands = [aband,]
                    else:
                        bands.append(aband)
                    if datarates is None:
                        datarates = [adatarate,]
                    else:
                        datarates.append(adatarate)

            return Stations(stations=(s for s in self.stations if any(n in s.networks for n in networks)
                                      and s.codename in the_network.station_codenames),
                            observing_bands=bands, max_datarates=datarates)

        return Stations(stations=(s for s in self.stations if any(n in s.networks for n in networks)),
                        observing_bands=self.observing_bands, max_datarates=self._bands.values())

    def filter_band(self, band: str) -> Stations:
        """Returns a new network which only contains the antennas that can observe
        at the given observing band.

        Input
            - band : str
                The observing band.

        Returns
            - Stations
                A new network object containing only the antennas that can observe at the given band.
        """
        return Stations(stations=(s for s in self.stations if band in s.bands), observing_bands=(band,),
                        max_datarates=self.max_datarate(band))

    def filter_antennas(self, codenames: Sequence[str]) -> Stations:
        """Returns a new network object which will only contain the stations
        defined by the given list of codenames. It will thus be a subset of the current
        network.

        Input
        - codenames : Sequence[str]
            List with the codenames of the stations that should be present in the new
            network.

        Returns
        - subnetwork : Stations
            A new Stations object containing only the defined stations.
        Exceptions
        - It may raise KeyError if one of the given codenames are not present
          among the current stations.
        """
        if not all([codename in self.station_codenames for codename in codenames]):
            unexpected_ant = set(codenames).difference(set(self.station_codenames))
            if len(unexpected_ant) == 1:
                rprint(f"[yellow bold]WARNING: The antenna with codename {unexpected_ant.pop()} is not present "
                       "in the current network.[/yellow bold]")
            else:
                rprint(f"[yellow bold]WARNING: The antennas with codenames {', '.join(list(unexpected_ant))} "
                       "are not present in the current network.[/yellow bold]")

        return Stations(stations=[s for s in self.stations if s.codename in codenames],
                        observing_bands=self.observing_bands, max_datarates=self._bands.values())

    @staticmethod
    def get_networks_from_configfile(filename: Optional[str] = None) -> dict[str, Stations]:
        """Reads a config file containing the different VLBI networks defined as a config parser file.
        Returns a dictionary with the nickname of the VLBI network as keys, and the information as
        keys in a second-order dict.

        Inputs
        - filename : str  OPTIONAL
            Path to the text file containing the information from all stations.
            If not provided, it reads the default station catalog file located in
                            data/network_catalog.inp
            Any other file should contain the same format (standard Python input config files),
            with the following fields per network (whose name would be provided as the name of
            section).
            - name - full name of the network.
            - default_antennas - comma-separated list with the codename of the antennas involved in the network.
            - max_datarate - integer number with the maximum datarate allowed in the network.
            - observing_bands - comma-separated list with the bands that can be observed (following
              the definition done in vlbiplanobs/freqsetups.py; e.g. 21, 18, etc, in cm).

        Returns
        - networks : dict[str, Stations]
            Returns a dictionary containing the differnet networks.
        """
        config = configparser.ConfigParser()
        if filename is None:
            with resources.as_file(resources.files("vlbiplanobs.data").joinpath("network_catalog.inp")) \
                                                                                as networks_catalog_path:
                config.read(networks_catalog_path)
        else:
            # With this approach it raises a FileNotFound exception.
            # Otherwise config will run smoothly and provide an empty list.
            config.read(open(filename, "r"))

        all_ants: Stations = Stations()
        networks: dict[str, Stations] = dict()
        for networkname in config.sections():
            temp: str = config[networkname]["max_datarate"]
            if "," in temp:
                max_dt: list[u.Quantity] | u.Quantity = [int(dt.strip())*u.Mb/u.s for dt in temp.split(",")]
            else:
                max_dt = int(temp)*u.Mb/u.s

            obs_bands = [b.strip() for b in config[networkname]["observing_bands"].split(",")]
            default_ant = [a.strip() for a in config[networkname]["default_antennas"].split(",")]
            assert all([ant in all_ants.station_codenames for ant in default_ant]), "The default antenna(s) " \
                   f"({', '.join([ant for ant in default_ant if ant not in all_ants])}) from '{networkname}' " \
                   "is not present in stations_catalog!"

            assert all([band in freqsetups.bands.keys() for band in obs_bands]), \
                   f"Observing band ({', '.join([b for b in obs_bands if b not in freqsetups.bands.keys()])}) " \
                   "not present in freqsetups.py!"

            if isinstance(max_dt, Sequence):
                temp = ', '.join([d for d in max_dt if int(d.value) not in freqsetups.data_rates.keys()])
                assert all([int(dr.value) in freqsetups.data_rates.keys() for dr in max_dt]), \
                       f"Data rate ({temp}) not present in freqsetups.py!"
            else:
                assert int(max_dt.value) in freqsetups.data_rates.keys(), \
                       f"Data rate ({max_dt}) not present in freqsetups.py!"

            antennas = [all_ants[[a.codename for a in all_ants].index(ant)] for ant in default_ant]
            assert len(antennas) > 0, f"No antennas found for the network {networkname}."

            networks[networkname] = Stations(stations=antennas, observing_bands=obs_bands, max_datarates=max_dt)

        return networks

    @staticmethod
    def _parse_station_from_configfile(stationname: str, station: dict) -> Station:
        """Parses the entry of a station from the config file and returns a Station object.

        Inputs
        - stationname : str
            The name of the station as defined in the header of the entry for
        - station : dict
            A dictionary containing the different information from a station, with the following
            fields per station (whose name would be provided as the name of section):
            - station - full name of the station.
            - code : codename assigned to the station. It must be unique (typically two letters).
            - network - main network to which it belongs to.
            - possible_networks - all networks the station can participate in (including 'network')
            - country - country where the station is located.
            - diameter - free format string with the diameter of the station
                        (optional more information in case of interferometers).
            - position = x, y, z (in meters). Geocentric position of the station.
            - mount = ALTAZ/EQUAT - type of the station mount.
            - ax1lim = (x0, x1) - Limits in degrees for the axis 1 of the mount.
            - ax2lim = (x0, x1) - Limits in degrees for the axis 2 of the mount.
            - ax1rate - Slewing speed in deg/s for the axis 1 of the mount.
            - ax2rate - Slewing speed in deg/s for the axis 2 of the mount.
            - ax1acc - Slewing acceleration in deg/s/s for the axis 1 of the mount.
            - ax2acc - Slewing acceleration in deg/s/s for the axis 2 of the mount.
            - real_time = yes/no - if the station can participate in real-time observations (e.g. e-EVN).
            - SEFD_**  - SEFD (in Jy units) of the station at the **cm band. If a given band is not present,
                        it is assumed that the station cannot observe it.
                        For example SEFD_21 = 500 means that the SEFD at 21cm is 500 Jy.
            - maxdatarate_** - maximum data rate of ** (in Mbps) that the station can achieve for the
              specify network name (e.g. EVN).
            - Any other attribute is accepted, but ignored in this code. That would easily allow future
              extensions of the code.

        Returns
            Station
        """
        temp = [float(i.strip()) for i in station["position"].split(",")]
        a_loc = coord.EarthLocation(u.Quantity(temp[0], u.m), u.Quantity(temp[1], u.m), u.Quantity(temp[2], u.m))
        # Getting the SEFD values for the bands
        does_real_time = True if station["real_time"] == "yes" else False
        sefds = {}
        max_dt: u.Quantity | dict[str, u.Quantity] | None = None
        for akey in station.keys():
            if "SEFD_" in akey.upper():
                sefds[f"{akey.upper().replace('SEFD_', '').strip()}"] = float(station[akey])*u.Jy
            if "maxdatarate" == akey.lower():
                max_dt = int(station[akey])*u.Mb/u.s
            elif "maxdatarate_" in akey.lower():
                if not isinstance(max_dt, dict):
                    max_dt = {}

                val = int(akey.removeprefix("maxdatarate_").strip())*u.Mb/u.s
                net = [n.strip() for n in station[akey].split(",")]
                for n in net:
                    max_dt[n] = val

        if all([key in station for key in ("mount", "ax1rate", "ax2rate", "ax1lim", "ax2lim")]):
            configs = {}
            try:
                for alim in ("ax1lim", "ax2lim"):
                    # Because they are the cable limits...
                    lims = [float(i) for i in station[alim].split(",")]
                    # easy check, if the range is >360, they can observe all azimuths
                    if (lims[1] - lims[0]) > 360:
                        configs[alim] = tuple((-1*u.deg, 361*u.deg))
                    else:
                        configs[alim] = tuple((lims[0]*u.deg, lims[1]*u.deg))

                if station["mount"] == 'EQUAT':
                    configs["ax1lim"] = tuple((ax1lim.value*u.hourangle for ax1lim in configs["ax1lim"]))

                for arate in ("ax1rate", "ax2rate"):
                    configs[arate] = u.Quantity(float(station[arate].strip()), u.deg/u.s)

                if all([key in station for key in ("ax1acc", "ax2acc")]):
                    for aacc in ("ax1acc", "ax2acc"):
                        configs[aacc] = u.Quantity(float(station[aacc].strip()), u.deg/u.s/u.s)

                    amount = Mount(MountType[station["mount"]], Axis(configs["ax1lim"],
                                                                     configs["ax1rate"],
                                                                     configs["ax1acc"]),
                                   Axis(configs["ax2lim"], configs["ax2rate"],
                                        station["ax2acc"]))
                else:
                    amount = Mount(MountType[station["mount"]], Axis(configs["ax1lim"],
                                   station["ax1rate"]), Axis(configs["ax2lim"],
                                   station["ax2rate"]))
            except ValueError:
                raise ValueError(f"when loading the data from antenna {station['station']}.")
        else:
            amount = None

        return Station(stationname, station["code"], tuple(station["networks"].split(",")),  # type: ignore
                       a_loc, sefds, station["station"], station["country"], station["diameter"],
                       does_real_time, amount, max_dt)

    @staticmethod
    def _get_stations_from_configfile(filename: Optional[str] = None,
                                      codenames: Optional[list[str]] = None) -> Generator[Station]:
        """Creates a Stations object (i.e. a network of stations) by reading the station
        information from an input file. Optionally, it allows to select only a subset of
        all stations in the file.

        Inputs
        - filename : str
            Path to the text file containing the information from all stations.
            If not provided, it reads the default station catalog file located in
                            data/stations_catalog.inp
            Any other file should contain the same format (standard Python input config files),
            with the following fields per station (whose name would be provided as the name of
            section).
            - station - full name of the station.
            - code : codename assigned to the station. It must be unique (typically two letters).
            - network - main network to which it belongs to.
            - possible_networks - all networks the station can participate in (including 'network')
            - country - country where the station is located.
            - diameter - free format string with the diameter of the station
                        (optional more information in case of interferometers).
            - position = x, y, z (in meters). Geocentric position of the station.
            - mount = ALTAZ/EQUAT - type of the station mount.
            - ax1lim = (x0, x1) - Limits in degrees for the axis 1 of the mount.
            - ax2lim = (x0, x1) - Limits in degrees for the axis 2 of the mount.
            - ax1rate - Slewing speed in deg/s for the axis 1 of the mount.
            - ax2rate - Slewing speed in deg/s for the axis 2 of the mount.
            - ax1acc - Slewing acceleration in deg/s/s for the axis 1 of the mount.
            - ax2acc - Slewing acceleration in deg/s/s for the axis 2 of the mount.
            - real_time = yes/no - if the station can participate in real-time observations (e.g. e-EVN).
            - SEFD_**  - SEFD (in Jy units) of the station at the **cm band. If a given band is not present,
                        it is assumed that the station cannot observe it.
                        For example SEFD_21 = 500 means that the SEFD at 21cm is 500 Jy.
            - maxdatarate_** - maximum data rate of ** (in Mbps) that the station can achieve for the
              specify network name (e.g. EVN).
            - Any other attribute is accepted, but ignored in this code. That would easily allow future
              extensions of the code.
        - codenames : list
            If you only want to select a subset of all stations available in the input file,
            here you can pass a list with the codenames of the stations that should be imported.

        Returns
        - network : Generator[Station]
            Returns a Stations object containing the specified stations.
        """
        config = configparser.ConfigParser()
        if filename is None:
            #             from importlib.resources import files
            # data_text = files('mypkg.data').joinpath('data1.txt').read_text()
            with resources.as_file(resources.files("vlbiplanobs.data").joinpath("stations_catalog.inp")) \
                                                                                 as stations_catalog_path:
                config.read(stations_catalog_path)
        else:
            # With this approach it raises a FileNotFound exception.
            # Otherwise config will run smoothly and provide an empty list.
            config.read(open(filename, "r"))

        for stationname in config.sections():
            if (codenames is None) or (config[stationname]["code"] in codenames):
                yield Stations._parse_station_from_configfile(stationname, config[stationname])  # type: ignore



"""Defines some of the possible values for the setup in a VLBI observation
"""
from typing import Union, Optional
from astropy import units as u

bands: dict[str, str] = {'92cm': '92 cm or 0.33 GHz', '49cm': '49 cm or 0.6 GHz',
                         '30cm': '30 cm or  1 GHz', '21cm': '21 cm or 1.4 GHz',
                         '18cm': '18 cm or 1.7 GHz', '13cm': '13 cm or 2.3 GHz',
                         '6cm': '6 cm or 5 GHz', '5cm': '5 cm or 6 GHz',
                         '3.6cm': '3.6 cm or 8.3 GHz', '2.5cm': '2.5 cm or 12 GHz',
                         '2cm': '2 cm or 15 GHz)',
                         '1.3cm': '1.3 cm or 23 GHz',  # '0.9cm': 'Ka band (0.9 cm or 33 GHz)',
                         '0.7cm': '0.7 cm or 43 GHz',
                         '0.348cm': '0.35 cm or 86 GHz',
                         '0.13cm': '0.13 cm or 230 GHz',
                         '0.09cm': '0.087 cm or 345 GHz'}

# from 4 Mbps to 32 Gbps
data_rates: dict[int, str] = {i: f"{i} Mbps" if i < 1e3 else f"{i/1000:.0f} Gbps"
                              for i in [2**j for j in range(2, 16)]}

# From 1 to 32 subbands (i.e. SPWs or IFs)
subbands: dict[int, str] = {2**i: f"{2**i} subbands" for i in range(6, -1, -1)}

# From 16 to 8192 spectral channels
channels: dict[int, str] = {2**i: f"{2**i} channels per subband" for i in range(14, 4, -1)}

# Full polarization meaning LL, RR, RL, LR; dual being RR and LL, and single just one of LL or RR.
polarizations: dict[int, str] = {4: '4 (Full polarization)', 2: '2 (dual polarization)',
                                 1: '1 (single polarization)'}

# Integration (averaging) times
inttimes: dict[Union[int, float], str] = {16: '16 s', 8: '8 s', 4: '4 s', 2: '2 s', 1: '1 s',
                                          0.5: '0.5 s', 0.25: '0.25 s', 0.125: '125 ms',
                                          0.06: '60 ms', 0.03: '30 ms', 0.015: '15 ms',
                                          0.001: '1 ms'}


def phaseref_cycle(band: str) -> Optional[u.Quantity]:
    """Returns the preferred phase referencing cycle for the given band for a standard VLBI observation.

    Input
        band :  str
            The observing band, e.g. '92cm'.
    Returns
        u.Quantity or None
            The preferred phase referencing cycle for the given band, e.g. 5*u.min at 6cm.
            Note that this cycle length means the time interval between two phase referencing scans,
            and thus it should be interpreted as the recommended length of a phasecal + target scans.
            It will return None if the observing frequency is > 60 GHz (wavelegnth < 0.5 cm), meaning
            no phase referencing is possible.
    """
    wavelength = float(band.replace('cm', ''))
    if wavelength > 15:
        return 7*u.min
    elif wavelength > 2:
        return 5*u.min
    elif wavelength > 0.5:
        return 2*u.min
    else:
        return None



import pytest
import numpy as np
# from astropy import coordinates as coord
from astropy import units as u
from astropy.time import Time
# from rich import print as rprint

from vlbiplanobs import observation as obs



def test_polarizations():
    assert obs.Polarization.SINGLE == obs.Polarization(1)
    assert obs.Polarization.SINGLE == obs.Polarization['SINGLE']
    assert obs.Polarization.DUAL == obs.Polarization(2)
    assert obs.Polarization.DUAL == obs.Polarization['DUAL']
    assert obs.Polarization.FULL == obs.Polarization(4)
    assert obs.Polarization.FULL == obs.Polarization['FULL']


def test_observation_init():
    o = obs.Observation()
    o.polarizations = 'dual'
    assert o.polarizations.value == 2
    o.polarizations = 4
    assert o.polarizations == obs.Polarization.FULL


    for tar_name, tar_coord in (('Target', '10h58m29.6s +81d33m58.8s'),
                    (['Tar1', 'Tar2'], ['10h58m29.6s +81d33m58.8s', '20h58m29.6s +71d33m58.8s'])):
        if isinstance(tar_name, list):
            target = [obs.Source(aname, acoord) for aname,acoord in zip(tar_name, tar_coord)]
        else:
            target = obs.Source(tar_name, tar_coord)

        o = obs.Observation(scans=obs.ScanBlock([obs.Scan(source=target)]))
        o.times = Time('2020-06-15 20:00', scale='utc') + np.arange(0, 720, 10)*u.min
        o.band = '18cm'
        o.datarate = 1024
        o.subbands = 8
        o.channels = 32
        o.polarizations = 2
        o.polarizations = 'full'
        with pytest.raises(ValueError):
            o.polarizations = 0
            o.polarizations = -5
            o.polarizations = 'efww'
            o.polarizations = ''
            o.polarizations = None

        o.inttime = 2

        if isinstance(tar_name, list):
            assert len(o.sources) == len(tar_name)
        else:
            assert len(o.sources) == 1

        # assert len(o.sources) == len(tar_name) if isinstance(tar_name, list) else 1
        # all_stations = stations.Network.get_stations_from_configfile()
        evn6 = ['Ef', 'Jb2', 'On', 'Hh', 'T6', 'Wb', 'Sv', 'Zc', 'Pa', 'Mp', 'Ho', 'Nl', 'Pt', 'Sc', 'Kp', 'Hn']
        o.stations_from_codenames(evn6)
        _ = o.elevations()
        return
        temp = o.is_visible()
        assert len(temp) == len(evn6) and len(evn6) == len(o.stations)
        # If multiple sources, then it will keep the structure of one ouput per source
        assert len(temp[evn6[0]]) == len(tar_name) if isinstance(tar_name, list) else 1
        # temp = obs.Observation.guest_times_for_source(target if not isinstance(target, list) else target[0],
        #                                               obs.Network.get_stations_from_configfile(codenames=evn6))
        # beam = o.synthesized_beam)
        temp = o.is_always_visible()
        assert len(temp['Ef']) == len(o.sources)
        assert temp['Ef']
        # _ = o.thermal_noise()
        # _ = o.get_uv_array()
        # _ = o.get_dirtymap(robust='natural')
        # _ = o.get_dirtymap(robust='uniform')
        # fig, ax = plt.subplots()

        # If no time is defined
        o.times = None
        with pytest.raises(ValueError):
            o.is_visible()

        o.with_networks(['EVN', 'eMERLIN'])
        assert 'Ef' in o.stations.codenames
        assert 'Hh' in o.stations.codenames
        assert 'Cm' in o.stations.codenames
        assert 'Jb2' in o.stations.codenames
        with pytest.raises(AssertionError):
            assert 'Nl' in o.stations.codenames
            assert 'Ku' in o.stations.codenames
            assert 'Pa' in o.stations.codenames






from typing import Optional, Union, Self, Sequence
from importlib import resources
import subprocess
import functools
# from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
import numpy as np
import yaml                # type: ignore
# from rich import print as rprint
import operator
from functools import reduce
import datetime as dt
from pathlib import Path
from enum import Enum, auto
from dataclasses import dataclass
from astropy import units as u
from astropy.time import Time
from astropy import coordinates as coord
# import polars as pl
from astroplan import FixedTarget


__all__ = ['SourceNotVisible', 'Source', 'SourceType', 'Scan', 'ScanBlock']

"""Defines an observation, which basically consist of a given network of stations,
observing a target source for a given time range and at an observing band.
"""


class SourceNotVisible(Exception):
    """Exception produced when a given target source cannot be observed for any
    antenna in the network.
    """
    pass


@dataclass
class FluxMeasurement:
    """Stores the fluxes related to the given source at a particular band (or frequency).

    - peak_flux : u.Quantity
        Refers to the peak brightness of the source. Meaning the peak expected in a map for the source in
        a VLBI map, i.e. the unresolved flux of the source.

    - flux_density : u.Quantity
        Refers to the total flux density of the source. Meaning the observed flux on the shortest baselines.
    """
    peak_flux: u.Quantity
    flux_density: u.Quantity


class SourceFlux(object):
    """Fluxes attributed to a given source.
    It provides the flux density (total flux) and peak flux (flux on the longest baselines)
    for a particular frequency.
    """

    def __init__(self, band_flux: dict[str, FluxMeasurement]):
        """Initializes a SourceFlux, which contains the peak flux and flux density of a given source at the
        given band.

        Input
            - band_flux : dict[str, FluxDensity]
                Dictionary of the form {band: FluxDensity}, with 'band' the different bands at which the
                flux measurements are referring to, and the corresponding FluxDensity object.
        """
        self._data: dict[str, FluxMeasurement] = band_flux

    def bands(self) -> tuple[str]:
        """Returns the bands at which there is flux information.
        """
        return tuple(self._data.keys())  # type: ignore

    def flux_density(self, band: str) -> FluxMeasurement:
        """Returns the flux density measurements associated to the source at the given band.

        Input
        - band : str
            The band at which the flux density measurements are referring to.

        Raises
            KeyError : if band is not available.
        """
        return self._data[band].flux_density

    def peak_flux(self, band: str) -> FluxMeasurement:
        """Returns the peak flux (brightness) measurements associated to the source at the given band.

        Input
        - band : str
            The band at which the flux density measurements are referring to.

        Raises
            KeyError : if band is not available.
        """
        return self._data[band].peak_flux

    def has_band(self, band: str) -> bool:
        """Returns if the band is present in the FluxMeasurement.
        """
        return band in self._data

    def __containts__(self, band: str) -> bool:
        return band in self._data

    def __getitem__(self, band: str) -> FluxMeasurement:
        return self._data[band]

    def __setitem__(self, band: str, flux: FluxMeasurement):
        self._data[band] = flux

    def add_band(self, band: str, flux: FluxMeasurement):
        """Adds a new measurement of the flux at a new frequency, or overwrites a previous one if the band exists.
        """
        if (not isinstance(band, str)) or (not isinstance(flux, FluxMeasurement)):
            raise TypeError("Expected 'band' to be a str and 'flux' to be a FluxMeasurement.")

        self._data[band] = flux


class SourceType(Enum):
    """Types of sources in a regular VLBI observation
    """
    TARGET = auto()
    PHASECAL = auto()
    FRINGEFINDER = auto()
    AMPLITUDECAL = auto()
    CHECKSOURCE = auto()
    POLCAL = auto()
    PULSAR = auto()
    UNKNOWN = auto()


class Source(FixedTarget):
    """Defines a target source located at some coordinates and with a given name.
    """

    def __init__(self, name: str,
                 coordinates: Optional[Union[str, coord.SkyCoord]] = None,
                 source_type: SourceType = SourceType.UNKNOWN,
                 flux: Optional[SourceFlux] = None,
                 notes: Optional[str] = None,
                 other_names: Optional[list[str]] = None, **kwargs):
        """Initializes a Source object.

        Inputs
        - name : str
            Name associated to the source.
        - coordinates : str or astropy.coordinates.SkyCoord [OPTIONAL]
            Coordinates of the target source in a str format recognized by
            astropy.coordinates.SkyCoord (e.g. XXhXXmXXs XXdXXmXXs).
            J2000 coordinates are assumed.
            If not provided, name must be a source name recognized by the RFC catalog or astroquery.
        - source_type : SourceType  [default = UNKNOWN]
            Defines the type of the source.
        - flux : SourceFlux  [default = None]
            Estimated flux density of the source at some given frequencies.
        - notes : str  [default = None]
            Some notes that you want to add for further information on the source.
        - other_names : list[str]  [default = None]
            A list of other possible names that the source may have.
        - kwargs
            keyword arguments to be passed to astropy.coordinates.SkyCoord() if needed.
            For example, the 'unit=' parameter.

        If both provided, the given coordinates will be used for the given source.

        It may raise:
        - NameResolveError: if the name is not recognized (and no coordinates are provided)
        - ValueError: if the coordinates have an unrecognized format.
        - AttributeError: if neither name or coordinates are provided, or name is empty.
        """
        if not isinstance(name, str):
            raise ValueError("'name' for Source needs to be a string (single source allowed).")

        if not isinstance(source_type, SourceType):
            raise ValueError("source_type must be a SourceType value.")

        if coordinates is None:
            coordinates = self.get_coordinates_from_name(name)

        super().__init__(coord.SkyCoord(coordinates, **kwargs), name)
        self._type = source_type
        self._flux = flux
        self._notes = notes
        self._other_names = other_names if other_names is not None else list()

    # @property
    # def coordinates(self) -> coord.SkyCoord
    #     return self.

    @property
    def other_names(self) -> list[str]:
        """Returns a list of other possible names to refer to this source.
        It may be None, if there are no other names.
        """
        return self._other_names

    @other_names.setter
    def other_names(self, other_names: list[str]):
        self._other_names = other_names

    @property
    def type(self) -> SourceType:
        return self._type

    @property
    def flux(self) -> Optional[SourceFlux]:
        """Estimated flux of the source.
        """
        return self._flux

    @property
    def notes(self) -> Optional[str]:
        """Some notes on the source
        """
        return self._notes

    @staticmethod
    def get_coordinates_from_name(src_name: str) -> coord.SkyCoord:
        """Returns the coordinates of a source by searching for them given the source name.
        First it searches in the RFC catalog, and if not found, in the ICRS catalogs.

        Inputs
        - src_name : str
            Name of the source to search for.

        Returns
        - src_coord : astropy.coordinates.SkyCoord
            The coordinates of the source, if found.

        It may raise
        - NameResolveError - if there is no connection or unable to find ICRS sources.
        - ValueError: if the coordinates have an unrecognized format.
        """
        try:
            return Source.get_rfc_coordinates(src_name)
        except ValueError:
            return coord.get_icrs_coordinates(src_name)

    @classmethod
    def source_from_name(cls, src_name: str) -> Self:
        """Returns a Source object by finding the coordinates from its name.
        It first searches wihtin the RFC catalog, and if not found, through the
        'astropy.coordinates.get_icrs_coordinates' function through the RFC catalogs.
        """
        return cls(src_name, coordinates=Source.get_coordinates_from_name(src_name))

    @classmethod
    def source_from_str(cls, src: str) -> Self:
        """Returns a Source object from the given string. This can be either the source name, and then it
        will retrieve the coordinates from the SIMBAD/NED/VizieR databases, if it's known, or the source
        coordinates (in either 'XXhXXmXXs XXdXXmXXs' or 'HH:MM:SS DD:MM:SS' formats).
        """
        if all([char in src for char in ('h', 'm', 'd', 's')]):
            return cls('target', coordinates=coord.SkyCoord(src), source_type=SourceType.TARGET)
        elif ':' in src:
            for char in ('h', 'm', 'd', 'm'):
                src = src.replace(':', char, 1)

            return cls('target', coordinates=coord.SkyCoord(src), source_type=SourceType.TARGET)
        else:
            return cls.source_from_name(src)

    @staticmethod
    def get_rfc_coordinates(src_name: str) -> coord.SkyCoord:
        """Returns the coordinates of the object by searching the provided name through the RFC catalog.
        """
        rfc_files = tuple((r.name for r in resources.files("vlbiplanobs.data").iterdir()
                           if r.is_file() and 'rfc' in r.name))
        assert len(rfc_files) > 0, "No RFC files found under the 'data' folder."

        with resources.as_file(resources.files("vlbiplanobs.data").joinpath(sorted(rfc_files)[-1])) as rfcfile:
            process = subprocess.run(["grep", src_name, rfcfile], capture_output=True, text=True)
        # process = subprocess.run(["grep", src_name, "./data/rfc_2021c_cat.txt"], capture_output=True, text=True)

        if process.returncode == 1:
            raise ValueError(f"The source {src_name} was not found in the RFC catalog.")
        elif process.returncode != 0:
            raise RuntimeError("A problem happened while parsing the RFC catalog file.")
        elif process.stdout.count('\n') != 1:
            raise ValueError(f"None or multiple coincidences for the source {src_name} in the RFC catalog.")

        temp = process.stdout.split()  # Fields:  IVS  name  J2000name  h m s d m s
        return coord.SkyCoord(f"{temp[3]}h{temp[4]}m{temp[5]}s {temp[6]}d{temp[7]}m{temp[8]}s")

    def sun_separation(self, times: Time) -> Sequence[u.Quantity]:
        """Returns the separation of the source to the Sun at the given epoch(s).

        Inputs
        - times : astropy.time.Time
            An array of times defining the duration of the observation. That is,
            the first time will define the start of the observation and the last one
            will be the end of the observation. Note that the higher time resolution
            in `times` the more precise values you will obtain for antenna source
            visibility or determination of the rms noise levels. However, that will
            also imply a longer computing time.

        Return
        - astropy.units.Quantity
            The separation between the source and the Sun at the given times.
        """
        return self.coord.transform_to(coord.GCRS(obstime=times)).separation(coord.get_sun(times))

    def sun_constraint(self, min_separation: u.Quantity, times: Optional[Time] = None) -> Time:
        """Checks if the Sun can be a restriction to observe the given source.
        This is defined as if the Sun is at a separation lower than 'min_separation' from the
        target in the sky at a given moment.

        Inputs
        - min_separation : astropy.units.Quantity
            Minimun separation allowed for the Sun to be near the source.
            One can make use of the default separations provided in `freqsetups.solar_separations`
            for the different observing bands.
        - times : astropy.time.Time  [OPTIONAL]
            An array of times to check the separation of the Sun. If not provided, it will compute
            the full (current) calendar year with a resolution of one day.

        Returns
        - astropy.time.Time
            A list of times when the Sun is closer than the minimum separation.
            It would be an empty list if this condition never meets.
        """
        if times is None:
            times = Time(f"{dt.datetime.now(tz=dt.timezone.utc).year}-01-01 00:00") + \
                    np.arange(0, 365, 1)*u.day

        sun_separation = self.sun_separation(times=times)
        if isinstance(sun_separation, list):
            return [times[sun_separation[i] < min_separation] for i in range(len(self.coord))]
        else:
            return times[sun_separation < min_separation]


class SourceCatalog:
    def __init__(self, personal_catalog: Optional[str] = None):
        self._blocks: dict[str, dict[str, ScanBlock]] = dict()
        # self._sources: dict[str, Source] = dict()
        # self._all_names: dict[str, str] = dict()
        # self._personal: set[str] = set()
        # self._catalog: set[str] = set()
        if personal_catalog is not None:
            self.read_personal_catalog(personal_catalog)

    @property
    def blocknames(self):
        return [bb for b in self._blocks.values() for bb in b.keys()]

    @property
    def blocks(self):
        return {bb_key: bb_value for b in self._blocks.values() for bb_key, bb_value in b.items()}

    @property
    def targets(self):
        return self._blocks['targets']

    @property
    def pulsars(self):
        return self._blocks['pulsars'] if 'pulsars' in self._blocks else None

    @property
    def ampcals(self):
        return self._blocks['ampcals'] if 'ampcals' in self._blocks else None

    @property
    def fringefinders(self):
        return self._blocks['fringefinders'] if 'fringefinders' in self._blocks else None

    @property
    def polcals(self):
        return self._blocks['polcals'] if 'polcals' in self._blocks else None

    @functools.cache
    def source_names(self, include_calibrators: bool = False) -> list[str]:
        """Returns the names of all sources in the database.
        """
        if include_calibrators:
            return list([s.name for b in self._blocks.values() for bs in b.values() for s in bs.sources()])

        return list([s.name for b in self._blocks['targets'].values() for s in b.sources()])

    @functools.cache
    def sources(self, include_calibrators: bool = False) -> dict[str, Source]:
        """Returns all sources.
        """
        if include_calibrators:
            return {s.name: s for b in self._blocks.values() for bs in b.values() for s in bs.sources()}

        return {s.name: s for b in self._blocks['targets'].values() for s in b.sources()}

    def __contains__(self, item: str):
        return item in self.blocknames

    def __getitem__(self, item: str):
        for key in self._blocks:
            if item in self._blocks[key]:
                return self._blocks[key][item]

        raise KeyError(f"The item {item} is not in the catalog.")

    def read_personal_catalog(self, path: str):
        """Reads the yaml file with the information of all sources that may be scheduled.
        It returns the catalog as a dictionary.

        Input
            path : str
                The path to the yaml file with the catalog of sources to be imported.
        """
        with open(path, 'r') as sources_yaml:
            catalog = yaml.safe_load(sources_yaml)
            for a_entry in catalog:
                if a_entry not in self._blocks:
                    self._blocks[a_entry] = dict()

                match a_entry:
                    case 'pulsars':
                        src_type = SourceType.PULSAR
                    case 'targets':
                        src_type = SourceType.TARGET
                    case 'ampcals':
                        src_type = SourceType.AMPLITUDECAL
                    # case 'checksources':
                    #     src_type = SourceType.CHECKSOURCE
                    # case 'phasecals':
                    #     src_type = SourceType.PHASECAL
                    case 'fringefinders':
                        src_type = SourceType.FRINGEFINDER
                    case 'polcals':
                        src_type = SourceType.POLCAL
                    case _:
                        src_type = SourceType.UNKNOWN

                for a_src in catalog[a_entry]:
                    src = catalog[a_entry][a_src]
                    scans = []
                    if 'phasecal' in src:
                        scans.append(Scan(source=Source(name=src['phasecal']['name'],
                                                        coordinates=src['phasecal']['coordinates'],
                                                        source_type=SourceType.PHASECAL),
                                          duration=float(src['phasecal']['duration'])*u.min
                                          if 'duration' in src['phasecal'] else None,
                                          every=int(src['phasecal']['every']) if 'every' in src['phasecal'] else -1))

                    if 'checkSource' in src:
                        scans.append(Scan(source=Source(name=src['checkSource']['name'],
                                                        coordinates=src['checkSource']['coordinates'],
                                                        source_type=SourceType.CHECKSOURCE),
                                          duration=float(src['checkSource']['duration'])*u.min
                                          if 'duration' in src['checkSource'] else None,
                                          every=int(src['checkSource']['every'])
                                          if 'every' in src['checkSource'] else -1))

                    scans.append(Scan(source=Source(name=src['name'] if 'name' in src else a_src,
                                                    coordinates=src['coordinates'],
                                                    source_type=src_type),
                                      duration=float(src['duration'])*u.min if 'duration' in src else None,
                                      every=int(src['every']) if 'every' in src else -1))

                    self._blocks[a_entry][a_src] = ScanBlock(scans)

    def read_rfc_catalog(self, path: Optional[Union[str, Path]] = None):
        """Reads the catalog yaml file with the information of all sources that may be scheduled.
        It returns the catalog as a dictionary.

        Input
            path : str
                The path to the yaml file with the catalog of sources to be imported.
        """
        # TODO: convert this to another module and use duckDB, should be much faster
        raise NotImplementedError
        if path is None:
            path = resources.as_file(resources.files("vlbiplanobs.data").joinpath("rfc_2021_cat.txt"))

        with open(path, 'rt') as stations_catalog_path:
            pass

        # This is old code

        if path is None:
            with resources.as_file(resources.files("vlbiplanobs.data").joinpath("rfc_2021_cat.txt")) \
                                                                              as stations_catalog_path:
                all_lines = open(stations_catalog_path, 'rt').readlines()
            # stations_catalog_path = 'data/rfc_2021c_cat.txt'
        else:
            with open(path, 'rt') as stations_catalog_path:
                all_lines = stations_catalog_path.readlines()

        for aline in all_lines:
            if aline.strip()[0] == '#':
                continue

            pars = aline.strip().split()
            if len(pars) != 25:
                raise ValueError(f"Expected 25 elements in a row but {len(pars)} found: {pars}")

            a_flux = {}
            for cm, apar in zip(('13', '6', '3.6', '2', '1.3'), (14, 16, 18, 20, 22)):
                if pars[apar] != '-1.00' and pars[apar].isnumeric():
                    a_flux[cm] = float(pars[apar])*u.Jy

            # if len(a_flux) > 0:
            #     grade = 9 if all([af > 0.7*u.Jy for af in a_flux.values()]) else 5
            # else:
            #     grade = 3

            self.add(Source(name=pars[2], grade=8 if pars[0] == 'C' else 5 if pars[0] == 'N' else 3,
                            other_names=[pars[1]], coordinates="{0}h{1}m{2}s {3}d{4}m{5}s".format(*pars[3:9]),
                            source_type=SourceType.PHASECAL, flux=a_flux, notes="RFC Source"), label='catalog')


@dataclass
class Scan:
    source: Source
    duration: u.Quantity = 10*u.min
    every: int = -1


class ScanBlock:
    """Defines a list of scans, each of them defined as a pointing to a given source during a given time.
    """

    def __init__(self, scans: list[Scan]):
        """Creates a block of scans, ideally a block to be observed with the target scans, phase-reference
        calibrator source (if needed), and check sources.

        For example, a block can consist on a single target scan (if this is a non phase-referencing observation).
        Then such scan will be repeated until fill the maximum observing time.

        On the contrary, in a phase-referencing observation a scan block can consist on one or multiple target
        scans, the associated phase-referencing scans, and possible check sources to be observed every certain
        number of target scans.

        If you want to schedule a fringe finder regularly during the observation, that is also possible.
        However, for isolated scans on these sources it will be preferred to be defined as different scan blocks.
        """
        if len(scans) == 0:
            raise ValueError("The scan block cannot contain an empty list of scans.")

        if not all([isinstance(s, Scan) for s in scans]):
            raise ValueError("All elements in the list of scans must be of Scan type.")

        self._scans = scans

    @property
    def scans(self) -> list[Scan]:
        return self._scans

    def has(self, source_type: SourceType) -> bool:
        """Returns if the given source type is included among the ones observed in the provided list of scans.
        """
        return any([s.source.type is source_type for s in self.scans])

    def sources(self, source_type: Optional[SourceType] = None) -> list[Source]:
        """Returns the sources with the given source types in this block.
        """
        if source_type is None:
            return [s.source for s in self.scans]

        return [s.source for s in self.scans if s.source.type is source_type]

    def sourcenames(self, source_type: Optional[SourceType] = None) -> list[str]:
        """Returns the source names with the given source types in this block.
        """
        if source_type is None:
            return [s.source.name for s in self.scans]

        return [s.source.name for s in self.scans if s.source.type is source_type]

    def scans_with_sources(self, source_type: SourceType) -> list[Scan]:
        """Returns the scans with the given source types in this block.
        """
        return [s for s in self._scans if s.source.type == source_type]

    def fill(self, max_duration: u.Quantity) -> list[Scan]:
        """Given the list of scans, returns the final arrangement of scans that fills the available time.
        This will follow the following conditions:
        - Repeats the target scan within the given time. If a `phasecal` is provided, then it will always
          bracket each target scan with this `phasecal`. If two or more phasecal are provided (e.g. P1, P2),
          then it will assume a multi phase-referencing technique for the target T:  P1 P2 T P1 P2 T P1 P2...
        - If some sources have the 'every = N > 0' constraint, then these will be scheduled every N cycles.
          For example, if no phasecal are provided and N = 3 for the C source, then: T T C T T C ...
          And if one phasecal is provided: P T P T P C P T P...

         NOTE: block scans should be easy!  This program is not prepared for the situation when you have
         multiple targets with multiple phase reference sources mixed.
        """
        # safety Checks
        if reduce(operator.add, [s.duration.to(u.min) for s in self.scans]) > max_duration:
            raise ValueError("The max_duration of the block cannot be shorter than the time of all single scans.")

        main_loop: list[Scan] = []

        # First it arranges the (phasecal)/target scans if exist
        if self.has(SourceType.PHASECAL):
            if not self.has(SourceType.TARGET):
                raise ValueError("If phase calibrator scans provided, then target scans must also be provided.")

            loop_duration = reduce(operator.add, [s.duration for s in self.scans_with_sources(SourceType.TARGET) +
                                                  self.scans_with_sources(SourceType.PHASECAL)])
            last_duration = reduce(operator.add, [s.duration for s in self.scans_with_sources(SourceType.PHASECAL)])
            # the second sum above is because the phase-referencing loop needs to be closed at the end.

            if any([s.every > -1 for s in self.scans
                    if s.source.type not in (SourceType.TARGET, SourceType.PHASECAL)]):
                # As there can be multiple sources to be observed every certain scans, better to do it incremental...
                # full_n_reps = (max_duration - last_duration)/(loop_duration*)
                booked_time, n_loop = 0*u.min, 1
                while True:
                    target_in_this_scan: list[Scan] = []
                    for a_scan in [s for s in self.scans if s.every > -1]:
                        if n_loop % a_scan.every == 0:
                            target_in_this_scan += [a_scan,]

                    if len(target_in_this_scan) == 0:
                        target_in_this_scan = self.scans_with_sources(SourceType.TARGET)

                    to_append = self.scans_with_sources(SourceType.PHASECAL) + target_in_this_scan
                    n_loop += 1
                    if reduce(operator.add, [a.duration for a in to_append]) + \
                       booked_time > max_duration - last_duration:
                        break

                    main_loop += to_append
                    booked_time += reduce(operator.add, [a.duration for a in to_append])
            else:
                main_loop += (self.scans_with_sources(SourceType.PHASECAL) +
                              self.scans_with_sources(SourceType.TARGET)) * \
                              int(max_duration.to(u.min).value // (loop_duration + last_duration).to(u.min).value)

            main_loop += self.scans_with_sources(SourceType.PHASECAL)
        else:
            target_duration = reduce(operator.add, [s.duration for s in self.scans_with_sources(SourceType.TARGET)])
            main_loop += self.scans_with_sources(SourceType.TARGET) * \
                int(max_duration.to(u.min).value // target_duration.to(u.min).value)

        return main_loop

    def __iter__(self):
        yield from self._scans



import pytest
from functools import reduce
import operator
import numpy as np
from astropy import coordinates as coord
from astropy import units as u
from astropy.time import Time
from rich import print as rprint
from vlbiplanobs import sources
# from vlbiplanobs import stations


def test_source():
    """Tests the Source class
    """
    with pytest.raises(TypeError):
        sources.Source()

    with pytest.raises(ValueError):
        sources.Source('a_name', 'coordinates')
        sources.Source('a_name', '00h00m00s xxhqqmees')
        sources.Source('a_name', '40h00m00s 10d10m10s')
        sources.Source('a_name', '00:00:00 100:00:00')
        sources.Source('a_name', '-30:00:00 10:00:00')
        # sources.Source('Cyg X-1')
        sources.Source(name='WQQWE')
        sources.Source('WQQWE')

    s1 = sources.Source('a_name', '10h20m10s 40d30m10s')
    s2 = sources.Source('a_name', '10:20:10 40:30:10', unit=(u.hourangle, u.deg))

    with pytest.raises(ValueError):
        sources.Source(['a_name', 'a_second_name'])
        sources.Source(['a_name', 'a_second_name'], ['10h20m10s 40d30m10s', '10h20m10s 40d30m10s'])
        sources.Source(coordinates=['10h20m10s 40d30m10s', '20h30m0s 20d20m10s'])

    try:
        s3 = sources.Source(name='Cyg X-1')
        assert s3.name == 'Cyg X-1'
        s3.coord.to_string('hmsdms')
        assert s3.coord.separation(coord.SkyCoord('19h58m21.6757344s +35d12m05.784516s')) < 0.01*u.arcsec
    except coord.name_resolve.NameResolveError:
        rprint("[yellow]WARNING: seems like no internet connection is available. "
               "Unable to resolve source names[/yellow]")

    assert s1.name == 'a_name'
    assert s2.name == 'a_name'
    assert s1.coord == s2.coord

    # checking that the distance to the Sun is right
    # times are spring equinox, summer solstice fall equinox
    times = Time(['2024-03-19 22:06', '2024-06-20 15:51', '2024-09-22 07:44'], scale='utc')
    s1 = sources.Source('equinox source', '0h0m0s 0d0m0s')
    # print(f"\n\nSeparations: {s1.sun_separation(times)}\n\n")
    assert all(np.abs(s1.sun_separation(times) - np.array([0.0, 90.0, 180])*u.deg) < 1*u.deg)
    # print(f"\n\nConstraints: {s1.sun_constraint(20*u.deg, times=times)}\n\n")
    assert len(s1.sun_constraint(20*u.deg, times=times)) == 1
    s1 = sources.Source('polar_source', '0h0m0s 90d0m0s')
    # print(f"\n\nSeparations: {s1.sun_separation(times)}\n\n")
    assert all(np.abs(s1.sun_separation(times[(0, 2),]) - np.array([90.0, 90.0])*u.deg) < 1*u.deg)

    # Playing with multiple sources
    # s1 = sources.Source(['s2', 's3'], ['10h20m10s 40d30m10s', '20h30m0s 20d20m10s'])
    # sep = s1.sun_separation(times)
    # assert len(sep) == 2
    # assert (len(sep[0]) == len(times)) and (len(sep[1]) == len(times))
    # s1.sun_constraint(20*u.deg)


def test_source_from_names():
    try:
        assert isinstance(sources.Source.source_from_name('Cyg X-1'), sources.Source)
        assert isinstance(sources.Source.get_coordinates_from_name('Cyg X-1'), coord.SkyCoord)

        src_name1 = '2358+390'
        src_name2 = 'J0000+3918'
        cor1 = sources.Source.get_coordinates_from_name(src_name1)
        cor2 = sources.Source.get_coordinates_from_name(src_name2)
        assert cor1 == cor2
        assert cor1.to_string('hmsdms') == '00h00m41.527583s +39d18m04.14836s'
        src1 = sources.Source.source_from_name(src_name1)
        src2 = sources.Source.source_from_name(src_name2)
        assert src1.coord == src2.coord
        assert src1.coord == cor2
    except coord.name_resolve.NameResolveError:
        rprint("[yellow]WARNING: seems like no internet connection is available. "
               "Unable to resolve source names[/yellow]")


def test_scan_block():
    target_source = sources.Source('target', '10h20m10s 40d30m10s', source_type=sources.SourceType.TARGET)
    pcal1_source = sources.Source('pcal1', '10h21m00s 40d31m00s', source_type=sources.SourceType.PHASECAL)
    pcal2_source = sources.Source('pcal2', '10h19m00s 40d29m00s', source_type=sources.SourceType.PHASECAL)
    check_source = sources.Source('check', '10h20m00s 40d31m00s', source_type=sources.SourceType.CHECKSOURCE)
    target_scan = sources.Scan(source=target_source, duration=3.5*u.min)
    pcal1_scan = sources.Scan(source=pcal1_source, duration=1.5*u.min)
    pcal2_scan = sources.Scan(source=pcal2_source, duration=1.5*u.min)
    check_scan = sources.Scan(source=check_source, duration=3*u.min, every=3)
    with pytest.raises(ValueError):
        sources.ScanBlock([])
        sources.ScanBlock(['lfla', 2])

    block = sources.ScanBlock([pcal1_scan, target_scan, pcal2_scan, check_scan])
    assert len(block.scans) == 4
    assert block.has(sources.SourceType.TARGET)
    assert block.has(sources.SourceType.PHASECAL)
    assert block.has(sources.SourceType.CHECKSOURCE)
    assert len(block.sources(sources.SourceType.TARGET)) == 1
    assert len(block.sources(sources.SourceType.PHASECAL)) == 2
    assert len(block.sources(sources.SourceType.CHECKSOURCE)) == 1
    assert len(block.sources(sources.SourceType.FRINGEFINDER)) == 0

    scans = block.fill(1*u.h)
    assert len(scans) == 29
    assert reduce(operator.add, [s.duration for s in scans]) <= 1*u.h
    assert all([scan.source.type is sources.SourceType.CHECKSOURCE for scan in scans[8::9]])



import sys
import argparse
from typing import Optional, Union
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
from astropy import units as u
from astropy.time import Time
from rich import print as rprint
from rich_argparse import RawTextRichHelpFormatter
import plotext as pltt
from vlbiplanobs import stations
from vlbiplanobs import observation as obs
from vlbiplanobs import sources
# from vlbiplanobs import scheduler

_STATIONS = stations.Stations()
_NETWORKS = _STATIONS.get_networks_from_configfile()


def get_stations(list_networks: Optional[list[str]] = None,
                 list_stations: Optional[list[str]] = None) -> stations.Stations:
    """Returns a VLBI array including the required stations.
    Each argument is a comma-separated list of names.

    Inputs
        list_networks : list[str] | None
            If you want to pick the default antennas participating in one of the
            known VLBI networks, then you can include directly the network name
            here.
        list_stations : list[str] | None
            If you want a particular list of stations, or adding some that are not
            in the default network, then you can quote them here, using either the
            station code names or their names.
    """
    stations = []
    try:
        if list_networks is not None:
            networks = [_NETWORKS[n] for n in list_networks]
            for n in networks:
                for s in n.station_codenames:
                    if s not in stations:
                        stations.append(s)
    except KeyError:
        unknown_networks: list = [n for n in list_networks if n not in _NETWORKS]  # type: ignore
        n_networks = len(unknown_networks)  # type: ignore
        rprint(f"[bold red]The network{'s' if n_networks > 1 else ''} {', '.join(unknown_networks)}"
               f" {'are' if n_networks > 1 else 'is'} not known.[/bold red]")
        sys.exit(1)

    try:
        if list_stations is not None:
            for s in list_stations:
                a_station = _STATIONS[s.strip()].codename
                if a_station not in stations:
                    stations.append(a_station)
    except KeyError:
        rprint(f"[bold red]The station {a_station} is not known.[/bold red]")
        sys.exit(1)

    return _STATIONS.filter_antennas(stations)


def summary(o: obs.Observation):
    """Prints the infromation for the given Observation, for testing purposes
    """
    rprint("\n[bold green]VLBI observation[/bold green]")
    rprint(f"To be conducted at {o.band.replace('cm', ' cm')} ", end='')
    if o.times is not None:
        rprint(f"from {o.times[0].strftime('%d %b %Y %H:%M')}--"
               f"{o.times[-1].strftime('%H:%M')} UTC")
    else:
        rprint("at unspecified times.")

    if o.duration is not None:
        rprint(f"With a total duration of {o.duration.to(u.h):.1}.")

    rprint(f"\n[bold green]Setup[/bold green]")
    if None not in (o.datarate, o.bandwidth, o.subbands):
        rprint(f"\nData rate of {o.datarate}, "
               f"{o.subbands} x {int(o.bandwidth.value/o.subbands)} "  # type: ignore
               f"{o.bandwidth.unit} subbands, with {o.channels} channels each, "  # type: ignore
               f"{o.polarizations} polarization.")
    else:
        rprint("[dim]No setup (data rate, bandwidth, number of subbands) specified[/dim]")

    rprint(f"\n[bold green]Stations ({len(o.stations)})[/bold green]: "
           f"{', '.join(o.stations.station_codenames)}")
    rprint("\n[bold green]Sources[/bold green]:")
    if o.scans is not None:
        for ablockname, ablock in o.scans.items():
            rprint(f"    - [dim]ScanBlock[/dim] '{ablockname}'\n      ")
            rprint('\n      '.join([s.name + ' [dim](' + s.coord.to_string('hmsdms') + ')[/dim]'
                                    for s in ablock.sources()]))

    print('\n')


def plot_visibility_tui(o: obs.Observation):
    """Show plots on the Terminal User Interface with the different sources and
    when they are visible within the observation.
    """
    if len(o.stations) == 0:
        rprint("[bold red]You need at least one station to observe![/bold red]")
        sys.exit(1)

    if o.times is None:
        rprint("[bold red]Searching for suitable GST range (with no defined observing time) not implemented yet[/bold red]")
        sys.exit(1)

    elevs = o.elevations()
    altaz = o.altaz()
    srcup = o.is_observable()
    srcupalways = o.is_always_observable()
    rprint(f"[bold]The blocks are observable for:[/bold]")
    for ablockname, antbool in srcup.items():
        rprint(f"    - '{ablockname}':", end='')
        if any(srcupalways[ablockname].values()):
            ant_can = [ant for ant, b in srcupalways[ablockname].items() if b]
            if len(ant_can) >= len(o.stations):
                rprint(f" [dim](always observable by {','.join(ant_can)})[/dim]")
            else:
                rprint(" [dim](always observable by everyone but ", end='')
                rprint(','.join([ant for ant in o.stations.station_codenames if ant not in ant_can]), end='')
                rprint(")[/dim]")
        else:
            rprint(' [dim](nobody can observe it all the time)[/dim]')

        for anti, ant in enumerate(antbool):
            rprint(f"        {ant:3}| {''.join(['◼︎' if b else ' ' for b in antbool[ant]])}")

        rprint(f"           |-{''.join(['-' for b in antbool[ant]])}|")
        rprint(f"           {o.times.datetime[0].strftime('%H:%M'):05}"
               f"{''.join([' ' for b in antbool[ant]][:-7])}"
               f"{o.times.datetime[-1].strftime('%H:%M'):05}")

        when_everyone = o.when_is_observable(mandatory_stations='all')[ablockname]
        if len(when_everyone) > 0:
            rprint("\n[bold]Everyone can observe the source at: [/bold]")
            rprint(', '.join([t1.strftime('%d %b %Y %H:%M')+'--'+t2.strftime('%H:%M') + ' UTC' for t1, t2 in when_everyone]))
        else:
            rprint("\nThe source cannot be observed by all stations at the same time.")

        min_stat = 3 if len(o.stations) > 3 else min(2, len(o.stations))
        rprint(f"[bold]Optimal visibility range (> {min_stat} antennas):[/bold] ")
        rprint(', '.join([t1.strftime('%d %b %Y %H:%M')+'--'+t2.strftime('%H:%M') \
                          + ' UTC' for t1, t2 in o.when_is_observable(min_stations=min_stat)[ablockname]]))


def plot_visibility(o: obs.Observation):
    """Show plots with the different sources and when they are visible within the
    observation
    """
    if o.times is None:
        rprint("[bold red]Searching for suitable GST range (with no defined observing time) "
               "not implemented yet[/bold red]")
        sys.exit(1)

    if o.scans is None:
        # rprint("No scans have been defined.")
        sys.exit(0)

    elevs = o.elevations()
    srcup = o.is_observable()
    srcupalways = o.is_always_observable()
    when = o.when_is_observable()
    for src in srcup:
        fig, ax = plt.subplots()
        for anti, ant in enumerate(srcup[src]):
            xs = o.times.mjd[srcup[src][ant]]
            ys = np.ones_like(xs) * anti
            targets = o.scans[src].sources(sources.SourceType.TARGET)
            if len(targets) > 0:
                colors = elevs[src][targets[0].name][ant][srcup[src][ant]].value
            else:
                colors = elevs[src][o.scans[src].sources()[0].name][ant][srcup[src][ant]].value

            points = np.array([xs, ys]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
            cmap = plt.cm.get_cmap('viridis')
            lc = LineCollection(segments, cmap=cmap, norm=plt.Normalize(0, 90))
            lc.set_array(colors)
            lc.set_linewidth(3)
            ax.add_collection(lc)
            # ax.set_yticks(np.arange(len(elevs[src])), labels=[ant for ant in elevs[src]])

        plt.colorbar(lc)
        plt.show()


def main(band: str, networks: Optional[list[str]] = None,
         stations: Optional[list[str]] = None,
         src_catalog: Optional[str] = None, targets: Optional[list[str]] = None,
         start_time: Optional[Time] = None,
         duration: Optional[u.Quantity | float] = None):
    """Planner for VLBI observations.

    Inputs
        targets: list[str]
            List of sources to be observed. Each entry can be:
            a) the name defining a block in the source catalog file (if provided),
            b) the coordinates of the source, in RA, DEC (J2000), as 'hh:mm:ss dd:mm:ss'
               or 'XXhXXmXXs XXdXXmXXs',
            c) the name of the source, if it is a known one so it can be found in the
               SIMBAD/NEW/VizieR databases.
            A mix of the previous ones can also be used for each entry.
    """
    if networks is None and stations is None:
        rprint("[bold red]You need to provide at least a VLBI network "
               "or a list of antennas that will participate in the observation.[/bold red]")
        sys.exit(1)

    if isinstance(duration, float):
        duration = duration * u.hour

    if start_time is not None and start_time.scale != 'utc':
        rprint("[bold red]The start time must be in UTC[/bold red]\n"
               "[red](use 'scale' when defining the Time object)[/red]")
        sys.exit(1)

    if src_catalog is not None:
        source_catalog = sources.SourceCatalog(src_catalog)
    else:
        source_catalog = None

    src2observe: dict[str, sources.ScanBlock] = {}
    if targets is not None:
        for target in targets:
            if (source_catalog is not None) and (target in source_catalog.blocknames):
                src2observe[target] = source_catalog[target]
            else:
                a_source = sources.Source.source_from_str(target)
                src2observe[a_source.name] = sources.ScanBlock([sources.Scan(a_source)])
    elif src_catalog is None:
        rprint("[bold red]Either a source catalog file or a list of targets must be "
               "provided (or both).[/bold red]")
        sys.exit(1)
    else:
        src2observe = source_catalog.blocks

    # I see a difference between different versions of Python (maybe astropy)?!!
    if isinstance(duration.to(u.min), float):
        duration_val = duration.to(u.min)
    else:
        duration_val = duration.to(u.min).value

    o = obs.Observation(band, get_stations(networks, stations), scans=src2observe,
                        times=start_time + np.arange(0, duration_val + 5, 10)*u.min
                        if start_time is not None else None, duration=duration,
                        # datarate=, subbands=, channels=, polarizations=, inttime=)
                        ontarget=0.6)
    summary(o)
    plot_visibility_tui(o)
    # plot_visibility(o)
    return o


def cli():
    usage = "%(prog)s [-h]  XXX"
    description = "EVN Observation Planner"
    parser = argparse.ArgumentParser(description=description, prog="planobs", usage=usage,
                                     formatter_class=RawTextRichHelpFormatter)
    parser.add_argument('-t', '--targets', type=str, default=None,
                        help="Source(s) to be observed. It can be either the coordinates of the source\n"
                        "(in 'hh:mm:ss dd:mm:ss' or 'XXhXXmXXs XXdXXmXXs' format), or the name of\nthe source, "
                        "if it is a known source in SIMBAD/NED/VizieR databases.\nOr if the personal source "
                        "catalog is defined by '--input', then this\nselects the block(s) defined in the file to use, "
                        "ignoring the rest of\nsources.")
    parser.add_argument('-i', '--input', type=str, default=None,
                        help="Input file containing the personal source catalog.\n"
                        "If provided, then the '--targets' will select the block(s) "
                        "defined in\nthis file.")
    parser.add_argument('-t1', '--starttime', type=str, default=None,
                        help="Start of the observation, with the foramt 'YYYY-MM-DD HH:MM' "
                        "in UTC.")
    parser.add_argument('-d', '--duration', type=str, default=None,
                        help="Total duration of the observation, in hours.")
    parser.add_argument('-n', '--network', type=str, help="Comma-separated list "
                        "of the VLBI network(s) that will participate in\nthe observation. "
                        "It will take the default stations in each network.\nIf 'stations' "
                        "is provided, then it will take both the default stations\nplus "
                        "the ones given in stations. [green]See '--list-antennas' to get a list.[/green]")
    parser.add_argument('-s', '--stations', type=str, help="Comma-separated list "
                        "of the antennas that will participate in the\nobservation. "
                        "You can use either antenna codenames or the standard name\n"
                        "as given in the catalogs. [green]See '--list-antennas' to get a list.[/green]")
    parser.add_argument('-b', '--band', type=str, help="Observing band, as defined "
                        "in the catalogs as 'XXcm', with 'XX' being\nthe wavelegnth in cm. "
                        "[green]See '--list-bands' to get a list.[/green]")
    # parser.add_argument('', type=, default=, help='')
    parser.add_argument('--list-antennas', action="store_true", default=False,
                        help="Writes the list of all antennas defined in PlanObs.")
    parser.add_argument('--list-bands', action="store_true", default=False,
                        help="Writes the list of all observing bands defined in PlanObs.")
    # parser.add_argument('', type=, default=, help='')
    subparser = parser.add_subparsers(dest='subparser', help="")
    parser.add_argument('--sched', default=None, type=str,
                        help="Produces a (SCHED) .key schedule file for "
                        "the observation with the\ngiven name.")
    parser.add_argument('--fringefinders', default='2', type=str,
                        help="Defines the fringe finder sources to be scheduled "
                        "in the observation.\nIt can be either a comma-separated "
                        "list of source names (as long as they\nappear in AstroGeo) "
                        "or a number, meaning how many scans should go on\nfringe "
                        "finders, and it will select the most suitable sources.")
    parser.add_argument('--polcal', action="store_true", default=False,
                        help="Requires polarization calibration for the observation.")
    parser.add_argument('--pulsar', default=None, type=str,
                        help="Sets to schedule at least a scan on a pulsar source. "
                        "If a number,\nit will select a pulsar from the personal "
                        "input source file (must be\nprovided!). If a name, "
                        "it will pick such pulsar.")

    args = parser.parse_args()

    if args.list_antennas:
        rprint("[bold]Available VLBI networks:[/bold]")
        for network_name, network in _NETWORKS.items():
            rprint(f"[bold]{network_name}[/bold]")
            rprint(f"  [dim]Default antennas: {', '.join(network.station_codenames)}[/dim]")

        rprint("\n[bold]All available antennas:[/bold]")
        for ant in _STATIONS:
            rprint(f"    {ant.codename} ({ant.name})  {ant.diameter} in {ant.country}")
            rprint(f"      [dim]Observes at {', '.join(ant.bands)}[/dim]")

    if args.list_bands:
        rprint("\n[bold]Available observing bands:[/bold]")
        for aband in obs.freqsetups.bands:
            rprint(f"[bold]{aband}[/bold] [dim]({obs.freqsetups.bands[aband]})[/dim]")
            rprint("[dim]  Observable with ", end='')
            rprint({', '.join([nn for nn, n in _NETWORKS.items() \
                                             if aband in n.observing_bands])}, end='')
            rorint("[/dim]")

    if args.list_antennas or args.list_bands:
        sys.exit(0)

    if args.band is None:
        parser.print_help()
        rprint("\n\n[bold red]The observing band (-b/--band) and --network and/or --stations are required.[/bold red]")
        exit(1)

    if args.band not in obs.freqsetups.bands:
        rprint(f"[bold red]The provided band ({args.band}) is not available"
               "[/bold red]\n[dim]Available values are: "
               f"{', '.join(obs.freqsetups.bands)}.[/dim]")
        sys.exit(1)

    if args.network is None and args.stations is None:
        rprint("[bold red]You need to provide at least a VLBI network "
               "or a list of antennas that will participate in the observation.[/bold red]")
        sys.exit(1)

    main(band=args.band,
         networks=[n.strip() for n in args.network.split(',')] if args.network else None,
         stations=[s.strip() for s in args.stations.split(',')] if args.stations else None,
         src_catalog=args.input,
         targets=args.targets.split(','), start_time=Time(args.starttime, scale='utc')
         if args.starttime else None, duration=args.duration*u.hour if args.duration else None)
    # o = obs.Observation(band=args.band, stations=get_stations(args.network, args.stations))


if __name__ == '__main__':
    cli()
